void CDOpen(char *fname);

FILE *ISOFILE;

#define ISODCL(from, to) (to - from + 1)

struct iso_volume_descriptor {
	char type[ISODCL(1,1)];
	char id[ISODCL(2,6)];
	char version[ISODCL(7,7)];
	char data[ISODCL(8,2048)];
};

#define ISO_VD_PRIMARY 1
#define ISO_VD_END 255

#define ISO_STANDARD_ID "CD001"

#define EL_TORITO_ID "EL TORITO SPECIFICATION"
#define EL_TORITO_ARCH_x86 0
#define EL_TORITO_ARCH_PPC 1
#define EL_TORITO_ARCH_MAC 2
#define EL_TORITO_BOOTABLE 0x88
#define EL_TORITO_MEDIA_NOEMUL 0
#define EL_TORITO_MEDIA_12FLOP  1
#define EL_TORITO_MEDIA_144FLOP 2
#define EL_TORITO_MEDIA_288FLOP 3
#define EL_TORITO_MEDIA_HD      4

struct iso_primary_descriptor {
	char type			[ISODCL (  1,   1)];
	char id				[ISODCL (  2,   6)];
	char version			[ISODCL (  7,   7)];
	char unused1			[ISODCL (  8,   8)];
	char system_id			[ISODCL (  9,  40)];
	char volume_id			[ISODCL ( 41,  72)];
	char unused2			[ISODCL ( 73,  80)];
	char volume_space_size		[ISODCL ( 81,  88)];
	char unused3			[ISODCL ( 89, 120)];
	char volume_set_size		[ISODCL (121, 124)];
	char volume_sequence_number	[ISODCL (125, 128)];
	char logical_block_size		[ISODCL (129, 132)];
	char path_table_size		[ISODCL (133, 140)];
	char type_l_path_table		[ISODCL (141, 144)];
	char opt_type_l_path_table	[ISODCL (145, 148)];
	char type_m_path_table		[ISODCL (149, 152)];
	char opt_type_m_path_table	[ISODCL (153, 156)];
	char root_directory_record	[ISODCL (157, 190)];
	char volume_set_id		[ISODCL (191, 318)];
	char publisher_id		[ISODCL (319, 446)];
	char preparer_id		[ISODCL (447, 574)];
	char application_id		[ISODCL (575, 702)];
	char copyright_file_id		[ISODCL (703, 739)];
	char abstract_file_id		[ISODCL (740, 776)];
	char bibliographic_file_id	[ISODCL (777, 813)];
	char creation_date		[ISODCL (814, 830)];
	char modification_date		[ISODCL (831, 847)];
	char expiration_date		[ISODCL (848, 864)];
	char effective_date		[ISODCL (865, 881)];
	char file_structure_version	[ISODCL (882, 882)];
	char unused4			[ISODCL (883, 883)];
	char application_data		[ISODCL (884, 1395)];
	char unused5			[ISODCL (1396, 2048)];
};

struct eltorito_boot_descriptor {
        char id	           		[ISODCL (  1,    1)];
	char id2			[ISODCL (  2,    6)];
	char version			[ISODCL (  7,    7)];
	char system_id			[ISODCL (  8,   39)];
	char unused2			[ISODCL ( 40,   71)];
        char bootcat_ptr                [ISODCL ( 72 ,  75)];
	char unused5			[ISODCL ( 76, 2048)];
};

struct eltorito_validation_entry {
        char headerid          		[ISODCL (  1,    1)];
	char arch			[ISODCL (  2,    2)];
	char pad1			[ISODCL (  3,    4)];
	char id  			[ISODCL (  5,   28)];
	char cksum			[ISODCL ( 29,   30)];
        char key1                       [ISODCL ( 31,   31)];
	char key2			[ISODCL ( 32,   32)];
};

struct eltorito_defaultboot_entry {
        char boot_id           		[ISODCL (  1,    1)];
	char boot_media			[ISODCL (  2,    2)];
	char loadseg			[ISODCL (  3,    4)];
	char arch  			[ISODCL (  5,    5)];
	char pad1			[ISODCL (  6,    6)];
        char nsect                      [ISODCL (  7,    8)];
	char bootoff			[ISODCL (  9,   12)];
        char pad2                       [ISODCL ( 13,   32)];
};


struct iso_path_table{
	unsigned char  name_len[2];
	char extent[4];
	char  parent[2];
	char name[1];
};

typedef struct  {
	unsigned char length			[ISODCL (1, 1)];
	char ext_attr_length			[ISODCL (2, 2)];
	char extent				[ISODCL (3, 10)];
	char size				[ISODCL (11, 18)];
	char date				[ISODCL (19, 25)];
	char flags				[ISODCL (26, 26)];
	char file_unit_size			[ISODCL (27, 27)];
	char interleave				[ISODCL (28, 28)];
	char volume_sequence_number		[ISODCL (29, 32)];
	unsigned char name_len			[ISODCL (33, 33)];
	char name[34];
} ISO_DIRECTORY_RECORD;


unsigned char dir_buffer[4096*2];

int get_int( char *p )
{
	return( ( p[0] & 0xff ) | (( p[1] & 0xff ) << 8 ) | (( p[2] & 0xff ) << 16 ) | (( p[3] & 0xff ) << 24 ) );
}

int get_short( char *p )
{
	return( ( p[0] & 0xff ) | (( p[1] & 0xff ) << 8 ) );
}

void mmssdd( int block, char *p )
{
	int m, s, d;
	
	block += 150;
	m = block / 4500;
	block = block - m * 4500;
	s = block / 75;
	d = block - s * 75;
	
	m = ( ( m / 10 ) << 4 ) | m % 10;
	s = ( ( s / 10 ) << 4 ) | s % 10;
	d = ( ( d / 10 ) << 4 ) | d % 10;	
	
	p[0] = BCD2INT( (char)m );
	p[1] = BCD2INT( (char)s );
	p[2] = BCD2INT( (char)d );
        p[3] = 0;
}

unsigned long csize, cstart;

unsigned char RAWISO[8192];
unsigned long PrevAdres=0xffffffff;

UINT8 *ISO_Read(UINT8 *cdloc)
{
            unsigned long adres;
            adres=(cdloc[0]*60*75)+(cdloc[1]*75)+cdloc[2];
            adres=adres-(75*2);

            if ((adres+1) != PrevAdres)
              fseek (ISOFILE, (adres*2352), SEEK_SET);

            PrevAdres=adres;
            fread (RAWISO, 2352, 1, ISOFILE);

            return((void *)RAWISO);
}

int loadISO(void)
{
	ISO_DIRECTORY_RECORD *root_dir;
        unsigned char loc[4];
        unsigned char *pos;

        ISOFILE=fopen (PSXloaded, "rb");
        ((UINT32*)loc)[0]=0x00100200;
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer, ((void*)pos)+24, 2048);
	root_dir = (ISO_DIRECTORY_RECORD *) &dir_buffer[156];

        mmssdd(get_int( root_dir->extent), loc);
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer, ((void*)pos+24), 2048);

        loc[2]=BCD2INT(loc[2])+1;
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer+2048, ((void*)pos+24), 2048);

        CDOpen ("SYSTEM.CNF;1");
        if (cstart==0) { return(1); }

        mmssdd(cstart, loc);
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer, ((void*)pos+24), csize);

 {
     unsigned char temp_name[256];
        sscanf((char*)dir_buffer, "BOOT = cdrom:%s;2",temp_name);
        if (temp_name[strlen(temp_name)-2]==';') temp_name[strlen(temp_name)-2] = 0;
        sprintf(PSXloaded,"%s",temp_name);
        psxload(PSXloaded);
 }

 return(0);
}

int loadISOEXE(char *name)
{
        EXE_HEADER tmpHead;
	ISO_DIRECTORY_RECORD *root_dir;
        unsigned char loc[4];
        unsigned char *pos;
        unsigned char tname[256];
        int csize=0, adres=0, i;

        ((UINT32*)loc)[0]=0x00100200;
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer, ((void*)pos)+24, 2048);
	root_dir = (ISO_DIRECTORY_RECORD *) &dir_buffer[156];

        mmssdd(get_int( root_dir->extent), loc);
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer, ((void*)pos+24), 2048);

        loc[2]=BCD2INT(loc[2])+1;
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer+2048, ((void*)pos+24), 2048);

        if (name[0]==0x5c) name=(void*)name+1;

        sprintf (tname, "%s;1", name);

        CDOpen (tname);
        if (cstart==0) { return(1); }

        mmssdd(cstart, loc);
        pos= ISO_Read(loc);
        memcpy (&tmpHead, ((void*)pos+24), sizeof(tmpHead) );

        csize=tmpHead.t_size>>11;
        adres=tmpHead.t_addr&0x01fffff;

        for (i=0; i<csize; i++)
         {
          loc[2]++;
          if (loc[2]>74)
           {
            loc[2]=0; loc[1]++;
            if (loc[1]>59) { loc[1]=0; loc[0]++;}
           }
          pos= ISO_Read(loc);
          memcpy (psxM+(adres+i*2048), ((void*)pos+24), 2048 );
         }
         CleanRec(tmpHead.t_addr&0x01fffff, tmpHead.t_size);
         rPC = tmpHead.pc0;
         reg[28] = tmpHead.gp0;
         reg[29] = tmpHead.saddr;

   asm("movl $_PSXcode, _NextFree");
   setmeml(PSXused,TENER,524288);

   ScanHLE(tmpHead.t_addr&0x01fffff, (tmpHead.t_size)+(tmpHead.t_addr&0x01fffff) );
 return(0);
}

void loadISOF(void)
{
        EXE_HEADER tmpHead;
	ISO_DIRECTORY_RECORD *root_dir;
        unsigned char loc[4];
        unsigned char *pos;
        unsigned char tname[256];
        int csize=0, adres=0, i;

if (psxM[(regA0&0x1fffff)+6]=='\\') i=7;
else i=6;

sprintf (tname, "%s", psxM+(regA0&0x1fffff)+i);

        ((UINT32*)loc)[0]=0x00100200;
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer, ((void*)pos)+24, 2048);
	root_dir = (ISO_DIRECTORY_RECORD *) &dir_buffer[156];

        mmssdd(get_int( root_dir->extent), loc);
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer, ((void*)pos+24), 2048);

        loc[2]=BCD2INT(loc[2])+1;
        pos= ISO_Read(loc);
        memcpy ((void*)dir_buffer+2048, ((void*)pos+24), 2048);

        CDOpen (tname);
        if (cstart==0) { return; }

        mmssdd(cstart, loc);
        pos= ISO_Read(loc);
        memcpy (&tmpHead, ((void*)pos+24), sizeof(tmpHead) );
        memcpy (psxM+(regA1&0x1fffff), ((void*)pos+24+0x10), 0x10 );

        csize=tmpHead.t_size>>11;
        adres=tmpHead.t_addr&0x01fffff;

        for (i=0; i<csize; i++)
         {
          loc[2]++;
          if (loc[2]>74)
           {
            loc[2]=0; loc[1]++;
            if (loc[1]>59) { loc[1]=0; loc[0]++;}
           }
          pos= ISO_Read(loc);
          memcpy (psxM+(adres+i*2048), ((void*)pos+24), 2048 );
         }

    regV0=1;
    rPC=regRA;
    CleanRec(tmpHead.t_addr&0x01fffff, tmpHead.t_size/4);

    ScanHLE(tmpHead.t_addr&0x01fffff, (tmpHead.t_size)+(tmpHead.t_addr&0x01fffff) );

 return;
}

void CDOpen(char *fname)
{
	int dir_rec_pos;
	ISO_DIRECTORY_RECORD *dir_rec;

	dir_rec_pos = 0;

 	while( dir_rec_pos < 4096 ) {
		dir_rec = (ISO_DIRECTORY_RECORD *) &dir_buffer[dir_rec_pos];
		if ( dir_rec->length[0] == 0 )
			break;
		dir_rec_pos += dir_rec->length[0];
		if ( ( dir_rec->name[0] == 0 ) || ( dir_rec->name[0] == 1 ) )
			continue;

                if ( dir_rec->flags[0] & 1 )
			continue;
		if ( strcmp( fname, dir_rec->name ) == 0 ) {
			cstart = get_int( dir_rec->extent );
			csize = get_int( dir_rec->size );
			return;
		}
	}
 	cstart = 0;
        return;
}	

